/*
 * phc -- the open source PHP compiler
 * See license/README.license for licensing information
 *
 * Main application module 
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <fstream>
#include <iostream>
#include "cmdline.h"
#include "PHP_parser.h"
#include "XML_parser.h"
#include "PHP_unparser.h"
#include "XML_unparser.h"
#include "DOT_unparser.h"
#include "Generate_IR.h"
#include "ast.h"
#include "process_ast.h"
#include "self_tests.h"

using namespace std;

struct gengetopt_args_info args_info;

void sigsegv_handler(int)
{
	fprintf(stderr, "Segmentation fault\n");
	exit(-1);
}

void sigabrt_handler(int)
{
	exit(-1);
}

int main(int argc, char** argv)
{
	AST_php_script* php_script = NULL;
	String* input_filename = NULL;

	// catch a seg fault - note this doesnt harm gdb, which will override this.
	// We do this so that tests will still continue if theres a seg fault
	signal(SIGSEGV, sigsegv_handler);
	signal(SIGABRT, sigabrt_handler);

	// Parse command line parameters 
	if(cmdline_parser(argc, argv, &args_info) != 0)
		exit(-1);

	// check for stdin
	bool use_stdin = false;
	bool use_xml_parser = false;
	if(args_info.inputs_num == 0)
	{
		use_stdin = true;
		input_filename = new String("-");
	}
	else
	{
		input_filename = new String(args_info.inputs[0]);

		// Check if the input file is an XML file
		int len = input_filename->length();
		if(len >= 4 && !strcmp(&(input_filename->c_str()[len-4]), ".xml"))
		{
			#ifdef HAVE_XERCES
				use_xml_parser = true;
				php_script = parse_ast_xml_file(input_filename);
			#else
				cout << "XML support not built in" << endl;
				exit(-1);
			#endif
		}

	}

	if (not use_xml_parser)
	{
		istream* input = NULL;
		ifstream* file = NULL;
		if(use_stdin)
		{
			input = &cin;
		}
		else
		{
			// Open input file
			file = new ifstream(args_info.inputs[0]);
			if(!file->is_open())
			{
				// TODO make an error code
				printf("Could not open '%s'\n", args_info.inputs[0]);
				exit(-1);
			}
			input_filename = new String(args_info.inputs[0]);
			input = file;
		}

		assert(input);
		assert(input_filename);

		// Compile
		PHP_parser parser(input, input_filename);

		if(args_info.dump_tokens_flag)
		{
			// run the lexer only
			while(parser.php_lexer->yylex());
		}
		else
		{
			if(parser.parse())
			{
				php_script = parser.php_script;
			}
		}

		if (not use_stdin)
		{
			assert(file);
			file->close();
		}
	}

	if(php_script != NULL)
	{
		if(args_info.test_linear_flag)
			test_linear(php_script);

		if(args_info.test_pre_vs_post_count_flag)
			test_pre_vs_post_count(php_script);

		if(args_info.test_cloning_flag)
			test_cloning(php_script);

		if(args_info.test_source_vs_semantic_values_flag)
			test_source_vs_semantic_values(php_script);
	
		if(args_info.test_xml_roundtrip_flag)
			test_xml_roundtrip(php_script);

		process_ast(php_script);

		if(args_info.generate_ir_flag)
		{
			Generate_IR generate_IR;
			php_script->visit(&generate_IR);
		}

		if(args_info.dump_php_flag)
		{
			PHP_unparser php_unparser;
			php_script->visit(&php_unparser);
		}

		if(args_info.dump_ast_xml_flag)
		{
			XML_unparser xml_unparser;
			php_script->visit(&xml_unparser);
		}

		if(args_info.dump_ast_dot_flag)
		{
			DOT_unparser dot_unparser;
			php_script->visit(&dot_unparser);
		}
	}

	return 0;
}
