/* 
 * phc -- the open source PHP compiler 
 * See license/README.license for licensing information
 *
 * Modify 'target' attribute on method invocations 
 * 
 * If the user does not specify a target for a method invocation, the parser
 * will set the corresponding field to NULL. However, target is not an optional
 * field; this transform will fix this. For methods defined in %MAIN%, target
 * will be set to %MAIN%. For other methods, target will be set to %STDLIB%.
 * Thus, if there is a name clash between a user-defined method and one in the
 * PHP library, we assume it is the users method that is called.
 *
 * The transform consists of two separate visitors: 
 *
 * Collect_method_names collects all method names that are defined in %MAIN%.
 * This visitor does not need to be called explicitely (it is called from
 * Invocation_targets)
 *
 * Invocation_targets then sets the 'target' attribute as described above.
 */

#ifndef PHC_INVOCATION_TARGETS
#define PHC_INVOCATION_TARGETS

#include "Tree_visitor.h"

class Collect_method_names : public Tree_visitor
{
public: 
	bool in_main;
	List<Token_method_name*> method_names;

public:
	Collect_method_names() : in_main(false) {}
	virtual ~Collect_method_names() {}

	void pre_class_def(AST_class_def* in)
	{
		in_main = *in->class_name->value == "%MAIN%";
	}
	
	void pre_signature(AST_signature* in)
	{
		if(in_main)
			method_names.push_back(in->method_name);
	}

	void children_method_invocation(AST_method_invocation* in)
	{
		// We avoid traversing the children of a method invocation
		// (because target has not yet been set)
	}
};

class Invocation_targets : public Tree_visitor
{ 
private: 
	Collect_method_names cmn;

public:
	Invocation_targets() {}
	virtual ~Invocation_targets() {}

	void pre_php_script(AST_php_script* in)
	{
		in->visit(&cmn);
	}

	void pre_method_invocation(AST_method_invocation* in)
	{
		if(in->target == NULL)
		{
			// TODO: Deal with reflection (impossible :-)
			Token_method_name* name = dynamic_cast<Token_method_name*>(in->method_name);
			if(name && defined_in_main(name))
				in->target = new Token_class_name(new String("%MAIN%"));
			else
				in->target = new Token_class_name(new String("%STDLIB%"));
		}
	}

private:
	bool defined_in_main(Token_method_name* name)
	{
		List<Token_method_name*>::const_iterator i;
		for(i = cmn.method_names.begin(); i != cmn.method_names.end(); i++)
		{
			if(*((*i)->value) == *name->value)
			{
				return true;
			}
		}
		return false;
	}
};

#endif // PHC_INVOCATION_TARGETS
