/* 
 * phc -- the open source PHP compiler 
 * See license/README.license for licensing information
 *
 * Replaces an include statement with the parsed contents of the included file.
 * 
 */

#ifndef PHC_PROCESS_INCLUDES
#define PHC_PROCESS_INCLUDES

#include "Tree_transform.h"
#include "Tree_visitor.h"
#include "ast.h"
#include "lib/error.h"
#include "PHP_unparser.h"
#include "XML_unparser.h"
#include "PHP_parser.h"
#include "Return_check.h"
#include <fstream>
#include <iostream>
#include <libgen.h>

/*
 * The include functionality is described in http://php.net/include/. 
 *
 * We support:
 * - include statements whose parameter must be a literal string
 * - the included file being in the current directory or the directory
 *	  containing the file being parsed
 * 
 * 
 * We don't support:
 * - return statements in the included file
 * - the '$val = include('my_file.php');' construct
 * - get_included_files()
 * - proper error messages. If it fails, we leave in the include statement
 *
 *
 * We intend to support in the future (but probably in the IR):
 * - include statements whose parameter can be derived using constant
 *   propegation or other dataflow analysis
 * - Run-time inclusion
 * - Remote compile-time or run-time inclusion
 * - return statements and the complications they entail
 */

/* Any variables available at that line in the calling file will be
 * available within the called file, from that point forward. */
// We put them in the current function

/* However, all functions and classes defined in the included file have
 * the global scope. */
// we put them in %MAIN% 

/* If the include occurs inside a function within the calling file, then all of
 * the code contained in the called file will behave as though it had been
 * defined inside that function. So, it will follow the variable scope of that
 * function. */
// Done: we put them in the current function

/* Because include() and require()  are special language constructs, you
 * must enclose them within a statement block if it's inside a conditional
 * block. */
// Done: This is handled automatically by the parser

/* $bar is the value 1 because the include was successful. Notice the
 * difference between the above examples. The first uses return() within the
 * included file while the other does not. If the file can't be included, FALSE
 * is returned and E_WARNING is issued. */
// Not done: we dont support this

/* Handling Returns: It is possible to execute a return()  statement
 * inside an included file in order to terminate processing in that file and
 * return to the script which called it. Also, it's possible to return values
 * from included files. You can take the value of the include call as you would
 * a normal function. This is not, however, possible when including remote
 * files unless the output of the remote file has valid PHP start and end tags
 * (as with any local file). You can declare the needed variables within those
 * tags and they will be introduced at whichever point the file was included.
 * */
// Not done: We'll handle returns in the IR, since the IR has goto, which these
// really need. Also, contructs like 
// '$a = do_something(do_something_else(include('myfile.php')))' are easy to
// support in the IR and nasty to support in the AST

/* If there are functions defined in the included file, they can be used in the
 * main file independent if they are before return() or after. If the file is
 * included twice, PHP 5 issues fatal error because functions were already
 * declared, while PHP 4 doesn't complain about functions defined after
 * return(). It is recommended to use include_once() instead of checking if the
 * file was already included and conditionally return inside the included file.
 * */
// Done: redefined functions signal an error

// With the _once, if there is an include_ or require_once, and there is an
// include anywhere which we cant include, fail all includes and issue a
// warning.
// Not done: just leaving this out. Include_once at your peril

// We could update get_included_files(), (say, by appending it with the file nmae of the included file), but that's a bit difficult to get right.
// Not done, a bit tricky


class PHP_parser;

class Process_includes : public Tree_transform
{
public:

	AST_php_script* current;

	AST_php_script* pre_php_script(AST_php_script* in)
	{
		current = in;
		return in;
	}


	// look for include statements
	void pre_eval_expr(AST_eval_expr* in, AST_statement_list* out)
	{
		bool pass_out_include_statement = true;

		// the filename is the only parameter of the include statement
		Token_string* token_filename = new Token_string(WILDCARD, WILDCARD);
		AST_actual_parameter* param = new AST_actual_parameter(false, token_filename);
		AST_actual_parameter_list* params = new AST_actual_parameter_list();
		params->push_back(param);

		// use an empty function name so we get it back
		Token_method_name* method_name = new Token_method_name(WILDCARD);


		// this will match the any function call with 1 string literal parameter,
		// and no lvalue
		AST_method_invocation *method = new AST_method_invocation(new Token_class_name(new String("%STDLIB%")), method_name, params);


		// check we have a matching function
		if	((in->expr->match(method)) and
			(*method_name->value == "include" or *method_name->value == "require"))
		{

			// search the include path, which, for now, is the directory
			// containing the currently parsing script, and current working
			// directory, in that order 
			ifstream* file = NULL;
			String* filename = NULL;

			// check in the same directory as the script
			// with stdin, we dont have a directory
			if (args_info.inputs_num > 0)
			{
				char* directory_name = strdup(args_info.inputs[0]);
				directory_name = dirname(directory_name);
				filename = new String(directory_name);
				filename->append("/");
				filename->append(token_filename->value->c_str());
				file = new ifstream(filename->c_str());
			}

			// if the file didnt open, try from the current running directory
			if(not file or not file->is_open())
			{
				file = new ifstream(token_filename->value->c_str());
				filename = token_filename->value;
			}



			if(not file->is_open())
			{
				error(ERR_INCLUDE_NO_FILE, in->get_line_number(), filename->c_str());
			}
			else
			{
				// parse the file
				PHP_parser parser(file, filename);
				if (not parser.parse())
				{
					error(ERR_INCLUDE_BAD_PARSE, in->get_line_number(), filename->c_str());
				}
				else
				{
					AST_php_script* php_script = parser.php_script;

					// Right now we deal with the statements to be returned.
					AST_class_def* main = php_script->get_class_def("%MAIN%");
					AST_method* run = main->get_method("%run%");

					// check for any return statements
					Return_check rc;
					for(AST_statement_list::const_iterator ci = run->statements->begin(); ci != run->statements->end(); ci++)
					{
						(*ci)->visit(&rc);
						if (rc.found) break;
					}

					if (rc.found)
					{
						error(ERR_INCLUDE_RETURN, in->get_line_number(), filename, rc.line_number);
					}
					else
					{
						// trying to keep this atomic. this says you can copy the results
						pass_out_include_statement = false;

						// copy the statements
						out->push_back_all(run->statements);


						// now move the classes to the currently running script
						for(	AST_class_def_list::const_iterator ci = php_script->class_defs->begin(); 
								ci != php_script->class_defs->end(); ci++)
						{
							// check the class isnt main
							if (*ci != main)
							{
								// check there isnt a class named this already
								if (current->get_class_def((*ci)->class_name->value->c_str()))
								{
									error(ERR_INCLUDE_CLASS_EXISTS, in->get_line_number(), (*ci)->class_name->value->c_str(), (*ci)->class_name->get_line_number());
								}
								else
								{
									current->class_defs->push_back(*ci);
								}
							}
						}

						// now deal with main's members and functions
						for(	AST_member_list::const_iterator mi =
								main->members->begin(); mi != main->members->end();
								mi++)
						{
							// check it's a method 
							AST_method* method = dynamic_cast<AST_method*>(*mi);
							if (method != NULL )
							{

								// check there isnt a method named this already and isnt %run%
								if (method != run)
								{
									if (current->get_class_def("%MAIN%")->get_method(method->signature->method_name->value->c_str()))
									{
										error(ERR_INCLUDE_METHOD_EXISTS, in->get_line_number(), method->signature->method_name->value->c_str(), method->get_line_number());
									}
									else
									{
										current->get_class_def("%MAIN%")->members->push_back(*mi);
									}
								}
							}
							else
							{
								current->get_class_def("%MAIN%")->members->push_back(*mi);
							}
						}
					}
					
				}

				file->close();
			}

		}

		// if the include statement did not work (and didnt die with an error,
		// just pass out the input
		if (pass_out_include_statement)
		{
			out->push_back(in);
		}
	}

};

#endif // PHC_PROCESS_INCLUDES
