/*
 * phc -- the open source PHP compiler
 * Convert the phc AST to XML format
 */

#ifndef PHC_XML_UNPARSER
#define PHC_XML_UNPARSER

#include "Visit_everything.h"
#include "lib/base64.h"
#include <iostream>
#include <typeinfo>
#include <cxxabi.h>

using namespace std;

class XML_unparser : public Visit_everything 
{
protected:
	ostream& os;
	int indent;

	void print_indent()
	{
		for(int i = 0; i < indent; i++)
			os << args_info.tab_arg;
	}

	const char* demangle(Object* obj)
	{
		int rv;
		const char* mangled = typeid(*obj).name();
		const char* demangled = abi::__cxa_demangle(mangled, NULL, NULL, &rv);
		if(rv == 0)
			return demangled;
		else
			return mangled;
	}

	bool needs_encoding(String* str)
	{
		String::const_iterator i;
		
		for(i = str->begin(); i != str->end(); i++)
		{
			if(*i < 32)
				return true;

			if(*i > 126)
				return true;

			if(*i == '<' || *i == '>' || *i == '&')
				return true;
		}

		return false;
	}

public:
	XML_unparser(ostream& os = cout) : os(os), indent(0) {}
	virtual ~XML_unparser() {}

	void pre_node(AST_node* in)
	{
		bool is_root = dynamic_cast<AST_php_script*>(in);
		
		if(is_root)
			os << "<?xml version=\"1.0\"?>" << endl;

		print_indent();
		os << "<" << demangle(in);
		
		if(is_root)
		{
			os << " xmlns=\"http://www.phpcompiler.org/phc-1.0\"";
			os << " xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"";
			os << " xsi:schemaLocation=\"http://www.phpcompiler.org/phc-1.0 generated/phc-1.0.xsd\"";
		}

		os << ">" << endl;
		indent++;

		if(in->attrs.size() == 0)
		{
			print_indent();
			os << "<attrs />" << endl;
		}
		else
		{
			print_indent();
			os << "<attrs>" << endl;
			indent++;

			map<string,Object*>::const_iterator i;
			for(i = in->attrs.begin(); i != in->attrs.end(); i++)
			{
				if((*i).first == "phc.line_number")
				{
					print_indent();
					os << "<attr key=\"phc.line_number\">" << in->get_line_number() << "</attr>" << endl;
				}
			}

			indent--;
			print_indent();
			os << "</attrs>" << endl;
		}
	}

	void post_node(AST_node* in)
	{
		indent--;
		print_indent();
		os << "</" << demangle(in) << ">" << endl;
	}

	void pre_identifier(AST_identifier* in)
	{
		String* value = in->get_value_as_string();

		print_indent();
		if(needs_encoding(value))
			os << "<value encoding=\"base64\">" << *base64_encode(value) << "</value>" << endl;
		else
			os << "<value>" << *value << "</value>" << endl;;

	}

	void pre_literal(AST_literal* in)
	{
		String* value = in->get_value_as_string();
		String* source_rep = in->get_source_rep();

		// Token_null does not have a value
		if(!dynamic_cast<Token_null*>(in))
		{
			print_indent();
			if(needs_encoding(value))
				os << "<value encoding=\"base64\">" << *base64_encode(value) << "</value>" << endl;
			else	
				os << "<value>" << *value << "</value>" << endl;
		}

		print_indent();
		if(needs_encoding(source_rep))
			os << "<source_rep encoding=\"base64\">" << *base64_encode(source_rep) << "</source_rep>" << endl;
		else	
			os << "<source_rep>" << *source_rep << "</source_rep>" << endl;
	}

public:
	void visit_null(char* type_id)
	{
		print_indent();
		os << "<" << type_id << " xsi:nil=\"true\" />" << endl;
	}

	void visit_marker(char* name, bool value)
	{
		print_indent();
		os << "<bool>" << (value ? "true" : "false") << "</bool>" << endl;
	}
};

#endif // PHC_XML_UNPARSER
