/*
 * phc -- the open source PHP compiler
 * See license/README.license for licensing information
 *
 * Series of self tests. Each of these tests is invoked by a particular
 * command line option; the test scripts in tests/ use these to do automated
 * testing of the various phc features.
 */

#include <iostream>
#include <sstream>
#include <typeinfo>
#include "self_tests.h"
#include "cmdline.h"
#include "Collect_all_pointers.h"
#include "Test_pre_vs_post_count.h"
#include "Get_source_and_semantic_values.h"
#include "PHP_unparser.h"
#include "XML_unparser.h"
#include "XML_parser.h"

using namespace std;

extern struct gengetopt_args_info args_info;

void
test_pre_vs_post_count(AST_php_script* script)
{
	Test_pre_vs_post_count t;
	script->visit(&t);
}

void
test_linear(AST_php_script* script)
{
	Collect_all_pointers cap;
	script->visit(&cap);

	if(cap.all_pointers.size() == cap.unique_pointers.size())
		printf("Success\n");
	else
		printf("Failure\n");
}

void
test_cloning(AST_php_script* origScript)
{
	bool success = true;

	// get the first script
	Collect_all_pointers origCap;
	origScript->visit(&origCap);

	// make a duplicate
	AST_php_script* dupScript = origScript->deep_clone();
	Collect_all_pointers dupCap;
	dupScript->visit(&dupCap);

	// Compare if the two scripts are the same
	if(!origScript->deep_equals(dupScript))
	{
		ostringstream orig_unparsed, dup_unparsed;
		PHP_unparser a(orig_unparsed), b(dup_unparsed);
		origScript->visit(&a);
		dupScript->visit(&b);

		printf("Failure\n");
		printf("Unparsing yields different results\n");
		cout << orig_unparsed.str();
		cout << dup_unparsed.str();
		success = false;
	}

	// Make sure the duplicated tree does not contain any pointers in the 
	// original tree
	list<AST_node*>::const_iterator i;
	for(i = dupCap.all_pointers.begin(); i != dupCap.all_pointers.end(); i++)
	{
		if(origCap.unique_pointers.find(*i) != origCap.unique_pointers.end())
		{
			printf("Failure\n");
			printf("Cloned tree contains pointers from original tree\n");
			success = false;
		}
	}
	
	if (success == true)
		printf("Success\n");
}

void test_source_vs_semantic_values(AST_php_script* script)
{
	Get_source_and_semantic_values gsasv;
	script->visit(&gsasv);
}

void test_xml_roundtrip(AST_php_script* script)
{
	#ifdef HAVE_XERCES
	ostringstream xml_output;
	XML_unparser xml_unparser(xml_output);
	script->visit(&xml_unparser);
	AST_php_script* reparsed = parse_ast_xml_buffer(new String(xml_output.str()));

	// TODO: change "match" to "deep_equals" once we store line numbers 
	// and comments in the XML file
	if(script->match(reparsed))
	{
		printf("Success\n");
	}
	else
	{
		printf("Failure\n");
		PHP_unparser php_unparser;
		script->visit(&php_unparser);
		reparsed->visit(&php_unparser);
	}
	#else
	cout << "XML support not built-in" << endl;
	#endif
}
